/* $USAGI: udpsend.c,v 1.14 2005/06/06 08:32:32 yoshfuji Exp $ */

/*
 * Copyright (C) 2001,2002,2003,2004 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <malloc.h>

#if defined(HAVE_CONFIG_H)
# include "../config.h"
#endif

#define	PACKET_STRING	"TestPacket"

int setup_ip6opt(struct cmsghdr *cmsg, int type, int optlen)
{
	unsigned char *buf, *p;
	if (optlen < 8 || optlen > 255 * 8 || optlen % 8) {
		fprintf(stderr, "invalid optlen: %d\n", optlen);
		return -1;
	}

	cmsg->cmsg_level = IPPROTO_IPV6;
	cmsg->cmsg_type = type;
	cmsg->cmsg_len = CMSG_LEN(optlen);

	p = CMSG_DATA(cmsg);

	memset(p, 0, optlen);

	p[1] = (optlen >> 3) - 1;

	p += 2;
	optlen -= 2;

	while (optlen > 0) {
		int len = optlen;

		if (len == 1) {
			/* Pad1 */
			p[0] = 0;
		} else {
			/* PadN */
			if (len > 257)
				len = 257;
			p[0] = 1;
			p[1] = len - 2;
		}

		p += len;
		optlen -= len;
	}

	return 0;
}

int main(int argc, char **argv) {
	const char *argv0 = argv[0];
	char *host, *serv;
	struct addrinfo hints, *res;
	int gai;
	int s;
	int len;
	char *buf, *cmsgbuf;
	int chunk_size = 100;
	int send_flags = 0;
	int pref_privacy = 0;
	int packet_size = 100;
	int c, usage = 0;
	int use_connect = 0;
	int send_dst = 0, send_hbh = 0;
	char *send_rthdr = NULL;
	int cmsg_space = 0;
	struct msghdr msg;
	struct cmsghdr *cmsg;

	while((c = getopt(argc, argv, 
				"c"
				"d:"
				"h:"
#if defined(MSG_MORE)
				"m:"
#endif
				"p:"
				"r:"
				"s:"
			  )) != EOF){
		switch (c) {
		case 'c':
			use_connect = 1;
			break;
		case 'd':
			send_dst = atoi(optarg);
			break;
		case 'h':
			send_hbh = atoi(optarg);
			break;
#if defined(MSG_MORE)
		case 'm':
			send_flags |= MSG_MORE;
			chunk_size = atoi(optarg);
			break;
#endif
		case 'p':
			pref_privacy = atoi(optarg);
			break;
#if defined(HAVE_INET6_RTHDR_SPACE)
		case 'r':
			send_rthdr = optarg;
			break;
#endif
		case 's':
			packet_size = atoi(optarg);
			break;
		default:
			usage = 1;
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (usage || argc < 2 || chunk_size <= 0 || packet_size <= 0) {
		printf("%s "
			"[-c] "
			"[-d dstopt_size] "
			"[-h hbhopt_size] "
#if defined(MSG_MORE)
			"[-m chunk_size] "
#endif
			"[-p pref_privacy] "
#if defined(HAVE_INET6_RTHDR_SPACE)
			"[-r addr,addr,...] "
#endif
			"[-s packet_size] "
			"host port\n", argv0);
		exit(1);
	}
	host = argv[0];
	serv = argv[1];

#if defined(MSG_MORE)
	if (!(send_flags & MSG_MORE))
		chunk_size = packet_size;
#else
	chunk_size = packet_size;
#endif
	buf = malloc(chunk_size);
	if (!buf) {
		perror("malloc");
		exit(1);
	}
	memset(buf, 0, chunk_size);
	snprintf(buf, chunk_size, "%s", PACKET_STRING);

	memset(&msg, 0, sizeof(msg));

	cmsg_space = (send_dst ? CMSG_SPACE(send_dst) : 0) + 
		     (send_hbh ? CMSG_SPACE(send_hbh) : 0);
	if (cmsg_space < send_dst ||
	    cmsg_space < send_hbh) {
		fprintf(stderr, "Invalid length\n");
		exit(1);
	}

#if defined(HAVE_INET6_RTHDR_SPACE)
	if (send_rthdr) {
		char *p;
		int count = 1;
		for (p = send_rthdr; *p; p++) {
			if (*p == ',')
				count++;
		}
		cmsg_space += inet6_rthdr_space(IPV6_RTHDR_TYPE_0, count);
	}
#endif

	if (cmsg_space) {
		cmsgbuf = malloc(cmsg_space);
		if (!cmsgbuf) {
			perror("malloc");
			exit(1);
		}

		msg.msg_control = cmsgbuf;
		msg.msg_controllen = cmsg_space;

#if defined(HAVE_INET6_RTHDR_SPACE)
		if (send_rthdr) {
			struct in6_addr addr;
			char *p, *p0, *buf = strdup(send_rthdr);

			if (!buf)
				exit(1);
			cmsg = inet6_rthdr_init(cmsgbuf, IPV6_RTHDR_TYPE_0);

			p = p0 = buf;
			do {
				if (*p == ',' || *p == '\0') {
					char c = *p;
					*p = '\0';
					if (inet_pton(AF_INET6, p0, &addr) <= 0)
						exit(1);
					inet6_rthdr_add(cmsg, &addr, IPV6_RTHDR_LOOSE);
					if (!c)
						break;
					p0 = p + 1;
				}
				p++;
			} while(1);
			inet6_rthdr_lasthop(cmsg, IPV6_RTHDR_LOOSE);

			free(buf);
			cmsg = CMSG_NXTHDR(&msg, cmsg);
		} else 
#endif
		{
			cmsg = CMSG_FIRSTHDR(&msg);
		}

		if (send_hbh) {
			if (setup_ip6opt(cmsg, IPV6_HOPOPTS, send_hbh) < 0)
				exit(1);
			cmsg = CMSG_NXTHDR(&msg, cmsg);
		}
		if (send_dst) {
			if (setup_ip6opt(cmsg, IPV6_DSTOPTS, send_dst) < 0)
				exit(1);
			cmsg = CMSG_NXTHDR(&msg, cmsg);
		}
	}
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	gai = getaddrinfo(host, serv, &hints, &res);
	if (gai) {
		printf("getaddrinfo() for %s port %s: %s\n",
			host, serv, gai_strerror(gai));
		exit(1);
	}

	s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (s < 0) {
		perror("socket");
		exit(1);
	}

	if (res->ai_family == PF_INET6 && pref_privacy) {
#ifdef IPV6_PRIVACY
		if (setsockopt(s, IPPROTO_IPV6, IPV6_PRIVACY, &pref_privacy, sizeof(pref_privacy)) < 0)
			perror("setsockopt(IPV6_PRIVACY)");
#else
		printf("IPV6_PRIVACY is not supported.\n");
#endif
	}	

	if (use_connect) {
		if (connect(s, res->ai_addr, res->ai_addrlen) < 0) {
			perror("connect");
			close(s);
			exit(1);
		}
	} else {
		msg.msg_name = res->ai_addr;
		msg.msg_namelen = res->ai_addrlen;
	}

	while (packet_size > 0) {
		struct iovec iov;

		if (packet_size <= chunk_size) {
			chunk_size = packet_size;
#if defined(MSG_MORE)
			send_flags &= ~MSG_MORE;
#endif
		}

		iov.iov_base = buf;
		iov.iov_len = chunk_size;

		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
#if 1
		len = sendmsg(s, &msg, send_flags);
		if (len < 0) {
			perror("sendmsg");
			close(s);
			exit(1);
		}
#else
		if (use_connect) {
			if ((len = send(s, buf, chunk_size, send_flags)) < 0) {
				perror("send");
				close(s);
				exit(1);
			}
		} else {
			if ((len = sendto(s, buf, chunk_size, send_flags,
						res->ai_addr, res->ai_addrlen)) < 0) {
				perror("sendto");
				close(s);
				exit(1);
			}
		}
#endif
		printf ("len = %d, flags = %08x, sent = %d\n", chunk_size, send_flags, len);
		packet_size -= chunk_size;
	}

	close(s);
	exit(0);
}

